package com.nubits.nubot.trading.wrappers;

import com.nubits.nubot.ALPClient.ErrorGeneratingALPRequestException;
import com.nubits.nubot.bot.Global;
import com.nubits.nubot.exchanges.ExchangeFacade;
import com.nubits.nubot.global.Constant;
import com.nubits.nubot.global.Settings;
import com.nubits.nubot.models.*;
import com.nubits.nubot.models.Currency;
import com.nubits.nubot.trading.ErrorManager;
import com.nubits.nubot.trading.ServiceInterface;
import com.nubits.nubot.trading.TradeInterface;
import com.nubits.nubot.trading.TradeUtils;
import com.nubits.nubot.trading.keys.ApiKeys;
import com.nubits.nubot.utils.Utils;
import org.apache.commons.codec.binary.Base64;
import org.json.simple.JSONArray;
import org.json.simple.JSONObject;
import org.json.simple.parser.JSONParser;
import org.json.simple.parser.ParseException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.crypto.Mac;
import javax.crypto.spec.SecretKeySpec;
import javax.net.ssl.HttpsURLConnection;
import java.io.*;
import java.net.MalformedURLException;
import java.net.ProtocolException;
import java.net.URL;
import java.net.URLEncoder;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

/**
 * Created by sammoth on 20/02/17.
 */
public class CryptopiaWrapper implements TradeInterface {
    private static final Logger LOG = LoggerFactory.getLogger(CryptopiaWrapper.class.getName());
    //Class fields
    private ApiKeys keys;
    protected CryptopiaService service;
    public final String EXCHANGE_NAME = ExchangeFacade.CRYPTOPIA;
    private final int TIME_OUT = 15000;
    private String checkConnectionUrl = "https://www.cryptopia.co.nz/api/";
    private final String SIGN_HASH_FUNCTION = "HmacSHA256";
    private final String ENCODING = "UTF-8";

    //handle concurrent requests
    private boolean isBusy;

    //Entry Points
    private final String API_BASE_URL = "https://www.cryptopia.co.nz/api/";
    private final String API_GET_BALANCE = "GetBalance";
    private final String API_GET_OPEN_ORDERS = "GetOpenOrders";
    private final String API_GET_TRADE_PAIRS = "GetTradePairs";
    private final String API_SUBMIT_TRADE = "SubmitTrade";
    private final String API_CANCEL = "CancelTrade";
    private final String API_GET_LAST_TRADES = "GetTradeHistory";

    //Tokens
    private final String TOKEN_BAD_RETURN = "No Connection With Exchange";
    private final String TOKEN_ERR = "Error";
    private final String TOKEN_SUCCESS = "Success";
    private final String TOKEN_DATA = "Data";
    private final String TOKEN_BAL_AVAIL = "Available";
    private final String TOKEN_BAL_HOLD = "HeldForTrades";
    private final String TOKEN_ORDER_ID = "OrderId";

    //Pool queue
    protected long lastRequest = 0;
    protected final long MIN_SPACING = 500;
    //Errors
    ErrorManager errors = new ErrorManager();

    public CryptopiaWrapper(ApiKeys keys) {
        this.keys = keys;
        service = new CryptopiaService(keys);
        setupErrors();
        this.setFree();
    }

    private void setupErrors() {
        errors.setExchangeName(this.EXCHANGE_NAME);
    }

    private ApiResponse getQuery(String url, String method, TreeMap<String, String> query_args, boolean needAuth, boolean isGet) {
        ApiResponse apiResponse = new ApiResponse();

        String queryResult = query(url, method, query_args, needAuth, isGet);
        if (queryResult == null) {
            apiResponse.setError(errors.nullReturnError);
            return apiResponse;
        }

        if (queryResult.equals(TOKEN_BAD_RETURN)) {
            apiResponse.setError(errors.noConnectionError);
            return apiResponse;
        }

        JSONParser parser = new JSONParser();

        try {
            JSONObject httpAnswerJson = (JSONObject) (parser.parse(queryResult));
            Boolean success = false;
            try {
                success = (Boolean) httpAnswerJson.get(TOKEN_SUCCESS);
            } catch (ClassCastException cce) {
                apiResponse.setError(errors.genericError);
            }

            if (!success) {
                String errorMessage = (String) httpAnswerJson.get(TOKEN_ERR);
                ApiError apiError = errors.apiReturnError;
                apiError.setDescription(errorMessage);
                apiResponse.setError(apiError);
            } else {
                apiResponse.setResponseObject(httpAnswerJson);
            }
        } catch (ClassCastException cce) {
            //if casting to a JSON object failed, try a JSON Array
            try {
                JSONArray httpAnswerJson = (JSONArray) (parser.parse(queryResult));
                apiResponse.setResponseObject(httpAnswerJson);
            } catch (ParseException pe) {
                LOG.error("httpResponse: " + queryResult + " \n" + pe.toString());
                apiResponse.setError(errors.parseError);
            }
        } catch (ParseException pe) {
            LOG.error("httpResponse: " + queryResult + " \n" + pe.toString());
            apiResponse.setError(errors.parseError);
            return apiResponse;
        }

        return apiResponse;
    }

    @Override
    public ApiResponse getAvailableBalances(CurrencyPair pair) {
        return getBalanceImpl(null, pair);
    }

    @Override
    public ApiResponse getAvailableBalance(Currency currency) {
        return getBalanceImpl(currency, null);
    }

    private ApiResponse getBalanceImpl(Currency currency, CurrencyPair pair) {
        ApiResponse apiResponse = new ApiResponse();
        boolean isGet = false;
        TreeMap<String, String> query_args = new TreeMap<>();

        String url = API_BASE_URL;
        String method = API_GET_BALANCE;

        ApiResponse response = getQuery(url, method, query_args, true, isGet);
        if (response.isPositive()) {
            JSONObject httpAnswerJson = (JSONObject) response.getResponseObject();
            JSONArray data = (JSONArray) httpAnswerJson.get(TOKEN_DATA);
            if (currency == null) {// pair was requested
                Amount NBTOnOrder = new Amount(0, pair.getOrderCurrency());
                Amount NBTAvailable = new Amount(0, pair.getOrderCurrency());
                Amount PegOnOrder = new Amount(0, pair.getPaymentCurrency());
                Amount PegAvailable = new Amount(0, pair.getPaymentCurrency());

                for (Iterator<JSONObject> wallet = data.iterator(); wallet.hasNext(); ) {
                    JSONObject thisWallet = wallet.next();
                    String thisCurrency = thisWallet.get("Symbol").toString();

                    if (pair.getOrderCurrency().getCode().toLowerCase().equals(thisCurrency.toLowerCase())) {
                        NBTOnOrder.setQuantity(Utils.getDouble(thisWallet.get(TOKEN_BAL_HOLD)));
                        NBTAvailable.setQuantity(Utils.getDouble(thisWallet.get(TOKEN_BAL_AVAIL)));
                    }
                    if (pair.getPaymentCurrency().getCode().toLowerCase().equals(thisCurrency.toLowerCase())) {
                        PegOnOrder.setQuantity(Utils.getDouble(thisWallet.get(TOKEN_BAL_HOLD)));
                        PegAvailable.setQuantity(Utils.getDouble(thisWallet.get(TOKEN_BAL_AVAIL)));
                    }
                }
                PairBalance balance = new PairBalance(PegAvailable, NBTAvailable, PegOnOrder, NBTOnOrder);
                apiResponse.setResponseObject(balance);

            } else { // just single currency
                Amount total = new Amount(0, currency);

                for (Iterator<JSONObject> wallet = data.iterator(); wallet.hasNext(); ) {
                    JSONObject thisWallet = wallet.next();
                    String thisCurrency = thisWallet.get("Symbol").toString();
                    if (thisCurrency.toLowerCase().equals(currency.getCode().toLowerCase())) {
                        total.setQuantity(Utils.getDouble(thisWallet.get(TOKEN_BAL_AVAIL)));
                    }
                }
                apiResponse.setResponseObject(total);
            }
        } else {
            apiResponse = response;
        }

        return apiResponse;
    }

    @Override
    public ApiResponse getLastPrice(CurrencyPair pair) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public ApiResponse sell(CurrencyPair pair, double amount, double rate) {
        return enterOrder(Constant.SELL, pair, amount, rate);
    }

    @Override
    public ApiResponse buy(CurrencyPair pair, double amount, double rate) {
        return enterOrder(Constant.BUY, pair, amount, rate);
    }

    public ApiResponse enterOrder(String type, CurrencyPair pair, double amount, double rate) {
        ApiResponse apiResponse = new ApiResponse();
        String order_id;
        boolean isGet = false;
        TreeMap<String, String> query_args = new TreeMap<>();

        query_args.put("Market", pair.toStringSepSpecial("/"));
        query_args.put("Rate", String.valueOf(rate));
        query_args.put("Amount", String.valueOf(amount));
        if (type.equals(Constant.BUY)) {
            query_args.put("Type", "Buy");
        } else {
            query_args.put("Type", "Sell");
        }

        String url = API_BASE_URL;
        String method = API_SUBMIT_TRADE;

        ApiResponse response = getQuery(url, method, query_args, true, isGet);
        if (response.isPositive()) {
            JSONObject httpAnswerJson = (JSONObject) response.getResponseObject();
            JSONObject dataJson = (JSONObject) httpAnswerJson.get(TOKEN_DATA);
            if (dataJson.containsKey(TOKEN_ORDER_ID)) {
                order_id = dataJson.get(TOKEN_ORDER_ID).toString();
                apiResponse.setResponseObject(order_id);
            } else {
                apiResponse.setError(errors.genericError);
            }
        } else {
            apiResponse = response;
        }

        return apiResponse;
    }

    @Override
    public ApiResponse getActiveOrders() {
        return getActiveOrdersImpl(null);
    }

    @Override
    public ApiResponse getActiveOrders(CurrencyPair pair) {
        return getActiveOrdersImpl(pair);
    }

    public ApiResponse getActiveOrdersImpl(CurrencyPair pair) {
        ApiResponse apiResponse = new ApiResponse();
        boolean isGet = false;
        TreeMap<String, String> query_args = new TreeMap<>();

        String url = API_BASE_URL;
        String method = API_GET_OPEN_ORDERS;

        ArrayList<Order> orderList = new ArrayList<Order>();

        ApiResponse response = getQuery(url, method, query_args, true, isGet);
        if (response.isPositive()) {
            JSONObject httpAnswerJson = (JSONObject) response.getResponseObject();
            JSONArray data = (JSONArray) httpAnswerJson.get(TOKEN_DATA);
            for (Iterator<JSONObject> order = data.iterator(); order.hasNext(); ) {
                JSONObject thisOrder = order.next();
                if (pair != null) {
                    if (CurrencyPair.getCurrencyPairFromString(thisOrder.get("Market").toString(), "/").equals(Global.options.getPair())) {
                        orderList.add(parseOrder(thisOrder));
                    }
                } else {
                    orderList.add(parseOrder(order.next()));
                }
            }
            apiResponse.setResponseObject(orderList);
        } else {
            apiResponse = response;
        }

        return apiResponse;
    }

    public Order parseOrder(JSONObject data) {
        Order order = new Order();

        //set the order id
        order.setId(data.get("OrderId").toString());
        CurrencyPair thisPair = CurrencyPair.getCurrencyPairFromString(data.get("Market").toString(), "/");
        order.setPair(thisPair);
        String type = data.get("Type").toString();
        if (type.equals("Sell")) {
            order.setType(Constant.SELL);
        } else {
            order.setType(Constant.BUY);
        }
        Amount price = new Amount(Utils.getDouble(data.get("Rate")), thisPair.getPaymentCurrency());
        order.setPrice(price);
        Amount amount = new Amount(Utils.getDouble(data.get("Amount")), thisPair.getOrderCurrency());
        order.setAmount(amount);
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd'T'kk:mm:ss.");
        Date date = null;
        try {
            date = sdf.parse(data.get("TimeStamp").toString());
        } catch (java.text.ParseException pe) {
            LOG.error(pe.toString());
        }
        if (date != null) {
            long ctime = date.getTime();
            Date insertDate = new Date(ctime);
            order.setInsertedDate(insertDate);
        }
        return order;
    }


    @Override
    public ApiResponse getOrderDetail(String orderID) {
        ApiResponse apiResponse = new ApiResponse();
        ApiResponse activeOrders = getActiveOrders();
        if (activeOrders.isPositive()) {
            ArrayList<Order> orderList = (ArrayList) activeOrders.getResponseObject();
            for (Iterator<Order> order = orderList.iterator(); order.hasNext(); ) {
                Order thisOrder = order.next();
                if (thisOrder.getId().equals(orderID)) {
                    apiResponse.setResponseObject(thisOrder);
                }
            }
        } else {
            apiResponse = activeOrders;
        }
        return apiResponse;
    }

    @Override
    public ApiResponse cancelOrder(String orderID, CurrencyPair pair) {
        ApiResponse apiResponse = new ApiResponse();
        boolean isGet = false;
        TreeMap<String, String> query_args = new TreeMap<>();

        query_args.put("Type", "Trade");
        query_args.put("OrderId", orderID);

        String url = API_BASE_URL;
        String method = API_CANCEL;

        ApiResponse response = getQuery(url, method, query_args, true, isGet);
        if (response.isPositive()) {
            JSONObject httpAnswerJson = (JSONObject) response.getResponseObject();
            JSONArray data = (JSONArray) httpAnswerJson.get(TOKEN_DATA);
            apiResponse.setResponseObject(false);
            for (Iterator<String> id = data.iterator(); id.hasNext(); ) {
                String thisId = id.next();
                if (thisId.equals(orderID)) {
                    apiResponse.setResponseObject(true);
                }
            }
        } else {
            apiResponse = response;
        }
        return apiResponse;
    }

    @Override
    public ApiResponse getTxFee() {
        return getTxFee(Global.options.getPair());
    }

    @Override
    public ApiResponse getTxFee(CurrencyPair pair) {
        ApiResponse apiResponse = new ApiResponse();
        boolean isGet = true;
        TreeMap<String, String> query_args = new TreeMap<>();

        String url = API_BASE_URL;
        String method = API_GET_TRADE_PAIRS;

        ApiResponse response = getQuery(url, method, query_args, true, isGet);
        if (response.isPositive()) {
            JSONObject httpAnswerJson = (JSONObject) response.getResponseObject();
            JSONArray data = (JSONArray) httpAnswerJson.get(TOKEN_DATA);
            for (Iterator<JSONObject> currencyDetail = data.iterator(); currencyDetail.hasNext(); ) {
                JSONObject thisCurrencyDetail = currencyDetail.next();
                if (thisCurrencyDetail.get("Label").toString().equals(pair.toStringSepSpecial("/").toUpperCase())) {
                    double txFee = Utils.getDouble(thisCurrencyDetail.get("TradeFee"));
                    apiResponse.setResponseObject(txFee);
                    return apiResponse;
                }
            }
        } else {
            apiResponse = response;
        }

        return apiResponse;
    }

    @Override
    public ApiResponse getLastTrades(CurrencyPair pair) {
        return getLastTradesImpl(pair, 0);
    }

    @Override
    public ApiResponse getLastTrades(CurrencyPair pair, long startTime) {
        return getLastTradesImpl(pair, startTime);
    }

    public ApiResponse getLastTradesImpl(CurrencyPair pair, long startTime) {
        ApiResponse apiResponse = new ApiResponse();
        boolean isGet = false;
        TreeMap<String, String> query_args = new TreeMap<>();

        String url = API_BASE_URL;
        String method = API_GET_LAST_TRADES;

        ArrayList<Trade> tradeList = new ArrayList<Trade>();

        query_args.put("Market", pair.toStringSepSpecial("/"));

        ApiResponse response = getQuery(url, method, query_args, true, isGet);
        if (response.isPositive()) {
            JSONObject httpAnswerJson = (JSONObject) response.getResponseObject();
            JSONArray data = (JSONArray) httpAnswerJson.get(TOKEN_DATA);
            for (Iterator<JSONObject> trade = data.iterator(); trade.hasNext(); ) {
                Trade thisTrade = parseTrade(trade.next());
                if (!thisTrade.getPair().equals(pair)) {
                    continue;
                }
                if (thisTrade.getDate().getTime() < startTime) {
                    continue;
                }
                tradeList.add(thisTrade);
            }
            apiResponse.setResponseObject(tradeList);
        } else {
            apiResponse = response;
        }

        return apiResponse;
    }

    public Trade parseTrade(JSONObject input) {
        Trade output = new Trade();
        /*
        {
           "TradeId": 23467,
           "TradePairId": 100,
           "Market": "DOT/BTC",
           "Type": "Buy",
           "Rate": 0.00000034,
           "Amount": 145.98000000,
           "Total": "0.00004963",
           "Fee": "0.98760000",
           "TimeStamp":"2014-12-07T20:04:05.3947572"
         }
         */
        output.setId(input.get("TradeId").toString());
        CurrencyPair pair = CurrencyPair.getCurrencyPairFromString(input.get("Market").toString(), "/");
        output.setPair(pair);
        output.setType(input.get("Type").toString().equals("Buy") ? Constant.BUY : Constant.SELL);
        Amount amount = new Amount(Utils.getDouble(input.get("Amount")), pair.getPaymentCurrency());
        output.setAmount(amount);
        Amount price = new Amount(Utils.getDouble(input.get("Rate")), pair.getOrderCurrency());
        output.setPrice(price);
        Amount fee = new Amount(Utils.getDouble(input.get("Fee")), pair.getOrderCurrency());
        output.setFee(fee);
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd'T'kk:mm:ss.");
        Date date = null;
        try {
            date = sdf.parse(input.get("TimeStamp").toString());
        } catch (java.text.ParseException pe) {
            LOG.error(pe.toString());
        }
        if (date != null) {
            long ctime = date.getTime();
            Date insertDate = new Date(ctime);
            output.setDate(insertDate);
        }
        return output;
    }

    @Override
    public ApiResponse isOrderActive(String id) {
        ApiResponse apiResponse = new ApiResponse();
        ApiResponse activeOrders = getActiveOrders();
        if (activeOrders.isPositive()) {
            ArrayList<Order> orderList = (ArrayList) activeOrders.getResponseObject();
            for (Iterator<Order> order = orderList.iterator(); order.hasNext(); ) {
                Order thisOrder = order.next();
                if (thisOrder.getId().equals(id)) {
                    if (thisOrder.getExecutedDate() == null) {
                        apiResponse.setResponseObject(true);
                    } else {
                        apiResponse.setResponseObject(false);
                    }
                }
            }
        } else {
            apiResponse = activeOrders;
        }
        return apiResponse;
    }

    @Override
    public ApiResponse getOrderBook(CurrencyPair pair) {
        throw new UnsupportedOperationException("CryptopiaWrapper.getOrderBook() not implemented yet.");
    }

    @Override
    public ApiResponse placeOrders(OrderBatch batch, CurrencyPair pair) {
        return TradeUtils.placeMultipleOrdersSequentiallyImplementation(batch, pair, 1000, this);
    }

    @Override
    public ApiResponse placeOrdersParallel(OrderBatch batch, CurrencyPair pair, ArrayList<ApiKeys> keys) {
        return TradeUtils.placeMultipleOrdersParallelImplementation(batch, pair, keys, this);
    }

    @Override
    public ApiResponse clearOrders(CurrencyPair pair) {
        ApiResponse apiResponse = new ApiResponse();
        boolean isGet = false;
        TreeMap<String, String> query_args = new TreeMap<>();

        query_args.put("Type", "All");

        String url = API_BASE_URL;
        String method = API_CANCEL;

        ApiResponse response = getQuery(url, method, query_args, true, isGet);
        if (response.isPositive()) {
            JSONObject httpAnswerJson = (JSONObject) response.getResponseObject();
            apiResponse.setResponseObject(true);
        } else {
            apiResponse = response;
        }
        return apiResponse;
    }

    @Override
    public ApiError getErrorByCode(int code) {
        return null;
    }

    @Override
    public String getUrlConnectionCheck() {
        return null;
    }

    @Override
    public String query(String base, String method, AbstractMap<String, String> args, boolean needAuth, boolean isGet) {
        String queryResult = TOKEN_BAD_RETURN; //Will return this string in case it fails
        if (this.isFree()) {
            this.setBusy();
            queryResult = service.executeQuery(base, method, args, needAuth, isGet);
            this.setFree();
        } else {
            //Another thread is probably executing a query. Init the retry procedure
            long sleeptime = Settings.RETRY_SLEEP_INCREMENT * 1;
            int counter = 0;
            long startTimeStamp = System.currentTimeMillis();
            LOG.debug(method + " blocked, another call is being processed ");
            boolean exit = false;
            do {
                counter++;
                sleeptime = counter * Settings.RETRY_SLEEP_INCREMENT; //Increase sleep time
                sleeptime += (int) (Math.random() * 200) - 100;// Add +- 100 ms random to facilitate competition
                LOG.debug("Retrying for the " + counter + " time. Sleep for " + sleeptime + "; Method=" + method);
                try {
                    Thread.sleep(sleeptime);
                } catch (InterruptedException e) {
                    LOG.error(e.toString());
                }

                //Try executing the call
                if (this.isFree()) {
                    LOG.debug("Finally the exchange is free, executing query after " + counter + " attempt. Method=" + method);
                    this.setBusy();
                    queryResult = service.executeQuery(base, method, args, needAuth, isGet);
                    this.setFree();
                    break; //Exit loop
                } else {
                    LOG.debug("Exchange still busy : " + counter + " .Will retry soon; Method=" + method);
                    exit = false;
                }
                if (System.currentTimeMillis() - startTimeStamp >= Settings.TIMEOUT_QUERY_RETRY) {
                    exit = true;
                    LOG.error("Method=" + method + " failed too many times and timed out. attempts = " + counter);
                }
            } while (!exit);
        }

        return queryResult;
    }

    @Override
    public SignedRequest getOpenOrdersRequest(CurrencyPair pair) throws ErrorGeneratingALPRequestException {
        return null;
    }

    @Override
    public void setKeys(ApiKeys keys) {

    }

    @Override
    public void setApiBaseUrl(String apiBaseUrl) {

    }

    @Override
    public boolean isBusy() {
        return isBusy;
    }

    @Override
    public boolean isFree() {
        return !isBusy;
    }

    @Override
    public void setBusy() {
        this.isBusy = true;
    }

    @Override
    public void setFree() {
        this.isBusy = false;
    }

    private class CryptopiaService implements ServiceInterface {

        protected ApiKeys keys;


        private final ExecutorService pool = Executors.newFixedThreadPool(5);

        public CryptopiaService(ApiKeys keys) {
            this.keys = keys;
        }

        @Override
        public String executeQuery(String base, String method, AbstractMap<String, String> args, boolean needAuth, boolean isGet) {
            HttpsURLConnection connection = null;
            boolean httpError = false;
            String output;
            int response = 200;
            String answer = null;
            URL queryUrl = null;

            JSONObject postData = new JSONObject();
            postData.putAll(args);
            try {
                queryUrl = new URL(base + method);
            } catch (MalformedURLException mal) {
                LOG.error(mal.toString());
                return null;
            }

            try {
                connection = (HttpsURLConnection) queryUrl.openConnection();
                connection.setRequestProperty("Content-type", "application/json");
                connection.setRequestProperty("User-Agent", Settings.APP_NAME);
                connection.setRequestProperty("Accept", "*/*");
                if (needAuth) {
                    connection.setRequestProperty("Authorization", getAuthHeader(queryUrl, postData));
                }
                connection.setRequestProperty("Connection", "close");

                connection.setDoOutput(true);
                connection.setDoInput(true);

                if (isGet) {
                    connection.setRequestMethod("GET");
                } else {
                    connection.setRequestMethod("POST");
                    DataOutputStream os = new DataOutputStream(connection.getOutputStream());
                    os.writeBytes(postData.toJSONString());
                    os.flush();
                    os.close();
                }
            } catch (ProtocolException pe) {
                LOG.error(pe.toString());
                return answer;
            } catch (IOException io) {
                LOG.error((io.toString()));
                return answer;
            }

            BufferedReader br = null;
            try {
                if (connection.getResponseCode() >= 400) {
                    httpError = true;
                    response = connection.getResponseCode();
                    answer = "";
                    br = new BufferedReader(new InputStreamReader(connection.getErrorStream()));
                } else {
                    answer = "";
                    br = new BufferedReader(new InputStreamReader(connection.getInputStream()));
                }
            } catch (IOException io) {
                LOG.error(io.toString());
                return answer;
            }

            if (httpError) {
                LOG.error("Query to : " + queryUrl
                        + "\nHTTP Response : " + Objects.toString(response));
            }

            try {
                while ((output = br.readLine()) != null) {
                    answer += output;
                }
            } catch (IOException io) {
                LOG.error(io.toString());
                return null;
            }

            connection.disconnect();
            connection = null;

            return answer;

        }

        private String hashPostData(String postData) {
            try {
                java.security.MessageDigest md = java.security.MessageDigest.getInstance("MD5");
                byte[] md5Hash = md.digest(postData.getBytes());
                byte[] base64Bytes = Base64.encodeBase64(md5Hash);
                return new String(base64Bytes);
            } catch (java.security.NoSuchAlgorithmException e) {
                LOG.error(e.toString());
            }
            return null;
        }

        private String getAuthHeader(URL url, JSONObject postParameters) {
            String nonce = Long.toString(System.currentTimeMillis());

            String requestParameter = this.keys.getApiKey() + "POST" +
                    URLEncoder.encode(url.toString()).toLowerCase() +
                    nonce +
                    hashPostData(postParameters.toJSONString());

            byte[] signedRequestParameter = signRequest(Base64.decodeBase64(this.keys.getPrivateKey()), requestParameter, SIGN_HASH_FUNCTION, ENCODING);
            byte[] base64Bytes = Base64.encodeBase64(signedRequestParameter);
            return "amx " + this.keys.getApiKey() + ":" + new String(base64Bytes) + ":" + nonce;
        }

        private byte[] signRequest(byte[] secret, String hash_data, String hashfFunction, String encoding) {
            byte[] sign = "".getBytes();
            try {
                Mac mac;
                SecretKeySpec key;
                // Create a new secret key
                key = new SecretKeySpec(secret, hashfFunction);
                // Create a new mac
                mac = Mac.getInstance(hashfFunction);
                // Init mac with key.
                mac.init(key);
                sign = mac.doFinal(hash_data.getBytes(encoding));
            } catch (UnsupportedEncodingException uee) {
                LOG.error("Unsupported encoding exception: " + uee.toString());
            } catch (NoSuchAlgorithmException nsae) {
                LOG.error("No such algorithm exception: " + nsae.toString());
            } catch (InvalidKeyException ike) {
                LOG.error("Invalid key exception: " + ike.toString());
            }
            return sign;
        }
    }
}

